/*
 *  Copyright (C) 2001 by Javier Achirica
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/* Standard headers */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>

#ifndef __FreeBSD__
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */
#include <linux/socket.h>	/* For AF_INET & struct sockaddr */
#include <linux/wireless.h>

#define WRQ_DATA(x) (x.u.data.pointer)
#else
#include <sys/socket.h>	
#include <net/if.h>
#include <net/if_var.h>
#include <net/ethernet.h>

#define WRQ_DATA(x) (x.ifr_data)
#endif

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

int socket_open(void)
{
	int sock;

	if ((sock=socket(AF_INET, SOCK_DGRAM, 0)) != -1)
		return sock;
#ifndef __FreeBSD__
	if ((sock=socket(AF_IPX, SOCK_DGRAM, 0)) != -1)
		return sock;
	if ((sock=socket(AF_AX25, SOCK_DGRAM, 0)) != -1)
		return sock;
#endif
	return socket(AF_APPLETALK, SOCK_DGRAM, 0);
}

static char *statsLabels[] = {
	"RxOverrun",
	"RxPlcpCrcErr",
	"RxPlcpFormatErr",
	"RxPlcpLengthErr",
	"RxMacCrcErr",
	"RxMacCrcOk",
	"RxWepErr",
	"RxWepOk",
	"RetryLong",
	"RetryShort",
	"MaxRetries",
	"NoAck",
	"NoCts",
	"RxAck",
	"RxCts",
	"TxAck",
	"TxRts",
	"TxCts",
	"TxMc",
	"TxBc",
	"TxUcFrags",
	"TxUcPackets",
	"TxBeacon",
	"RxBeacon",
	"TxSinColl",
	"TxMulColl",
	"DefersNo",
	"DefersProt",
	"DefersEngy",
	"DupFram",
	"RxFragDisc",
	"TxAged",
	"RxAged",
	"LostSync-MaxRetry",
	"LostSync-MissedBeacons",
	"LostSync-ArlExceeded",
	"LostSync-Deauth",
	"LostSync-Disassoced",
	"LostSync-TsfTiming",
	"HostTxMc",
	"HostTxBc",
	"HostTxUc",
	"HostTxFail",
	"HostRxMc",
	"HostRxBc",
	"HostRxUc",
	"HostRxDiscard",
	"HmacTxMc",
	"HmacTxBc",
	"HmacTxUc",
	"HmacTxFail",
	"HmacRxMc",
	"HmacRxBc",
	"HmacRxUc",
	"HmacRxDiscard",
	"HmacRxAccepted",
	"SsidMismatch",
	"ApMismatch",
	"RatesMismatch",
	"AuthReject",
	"AuthTimeout",
	"AssocReject",
	"AssocTimeout",
	"ReasonOutsideTable",
	"ReasonStatus1",
	"ReasonStatus2",
	"ReasonStatus3",
	"ReasonStatus4",
	"ReasonStatus5",
	"ReasonStatus6",
	"ReasonStatus7",
	"ReasonStatus8",
	"ReasonStatus9",
	"ReasonStatus10",
	"ReasonStatus11",
	"ReasonStatus12",
	"ReasonStatus13",
	"ReasonStatus14",
	"ReasonStatus15",
	"ReasonStatus16",
	"ReasonStatus17",
	"ReasonStatus18",
	"ReasonStatus19",
	"RxMan",
	"TxMan",
	"RxRefresh",
	"TxRefresh",
	"RxPoll",
	"TxPoll",
	"HostRetries",
	"LostSync-HostReq",
	"HostTxBytes",
	"HostRxBytes",
	"ElapsedUsec",
	"ElapsedSec",
	"LostSyncBetterAP",
	"PrivacyMismatch",
	"Jammed",
	"DiscRxNotWepped",
	"PhyEleMismatch",
	(char*)-1 };

static char *country[] = {
	"North America/Canada",
	"Europe (ETSI)",
	"Japan",
	"Spain",
	"France",
	"Belgium",
	"Israel",
	"Canada-Outdoor" };

/* This structure came from an email sent to me from an engineer at
   aironet for inclusion into this driver */
typedef struct {
	u16 len;
	u16 kindex;
	u8 mac[6];
	u16 klen;
	u8 key[16];  /* 40-bit keys */
} WepKeyRid;

/* These structures are from the Aironet's PC4500 Developers Manual */
typedef struct {
	u16 len;
	u8 ssid[32];
} Ssid;

typedef struct {
	u16 len;
	Ssid ssids[3];
} SsidRid;

typedef struct {
        u16 len;
        u16 modulation;
#define MOD_DEFAULT 0
#define MOD_CCK 1
#define MOD_MOK 2
} ModulationRid;

typedef struct {
	u16 cmd;
	u16 parm0;
	u16 parm1;
	u16 parm2;
} Cmd;

typedef struct {
	u16 status;
	u16 rsp0;
	u16 rsp1;
	u16 rsp2;
} Resp;

typedef struct {
	u16 len; /* sizeof(ConfigRid) */
	u16 opmode; /* operating mode */
#define MODE_STA_IBSS 0
#define MODE_STA_ESS 1
#define MODE_AP 2
#define MODE_AP_RPTR 3
#define MODE_ETHERNET_HOST (0<<8) /* rx payloads converted */
#define MODE_LLC_HOST (1<<8) /* rx payloads left as is */
#define MODE_AIRONET_EXTEND (1<<9) /* enable Aironet extenstions */
#define MODE_AP_INTERFACE (1<<10) /* enable ap interface extensions */
	u16 rmode; /* receive mode */
#define RXMODE_BC_MC_ADDR 0
#define RXMODE_BC_ADDR 1 /* ignore multicasts */
#define RXMODE_ADDR 2 /* ignore multicast and broadcast */
#define RXMODE_RFMON 3 /* wireless monitor mode */
#define RXMODE_RFMON_ANYBSS 4
#define RXMODE_LANMON 5 /* lan style monitor -- data packets only */
#define RXMODE_DISABLE_802_3_HEADER (1<<8) /* disables 802.3 header on rx */
	u16 fragThresh;
	u16 rtsThres;
	u8 macAddr[6];
	u8 rates[8];
	u16 shortRetryLimit;
	u16 longRetryLimit;
	u16 txLifetime; /* in kusec */
	u16 rxLifetime; /* in kusec */
	u16 stationary;
	u16 ordering;
	u16 u16deviceType; /* for overriding device type */
	u16 _reserved1[5];
	/*---------- Scanning/Associating ----------*/
	u16 scanMode;
#define SCANMODE_ACTIVE 0
#define SCANMODE_PASSIVE 1
#define SCANMODE_AIROSCAN 2
	u16 probeDelay; /* in kusec */
	u16 probeEnergyTimeout; /* in kusec */
        u16 probeResponseTimeout;
	u16 beaconListenTimeout;
	u16 joinNetTimeout;
	u16 authTimeout;
	u16 authType;
#define AUTH_OPEN 0x1
#define AUTH_ENCRYPT 0x101
#define AUTH_SHAREDKEY 0x102
	u16 associationTimeout;
	u16 specifiedApTimeout;
	u16 offlineScanInterval;
	u16 offlineScanDuration;
	u16 linkLossDelay;
	u16 maxBeaconLostTime;
	u16 refreshInterval;
#define DISABLE_REFRESH 0xFFFF
	u16 _reserved1a[1];
	/*---------- Power save operation ----------*/
	u16 powerSaveMode;
#define POWERSAVE_CAM 0
#define POWERSAVE_PSP 1
#define POWERSAVE_PSPCAM 2
	u16 sleepForDtims;
	u16 listenInterval;
	u16 fastListenInterval;
	u16 listenDecay;
	u16 fastListenDelay;
	u16 _reserved2[2];
	/*---------- Ap/Ibss config items ----------*/
	u16 beaconPeriod;
	u16 atimDuration;
	u16 hopPeriod;
	u16 channelSet;
	u16 channel;
	u16 dtimPeriod;
	u16 _reserved3[2];
	/*---------- Radio configuration ----------*/
	u16 radioType;
#define RADIOTYPE_DEFAULT 0
#define RADIOTYPE_802_11 1
#define RADIOTYPE_LEGACY 2
	u8 rxDiversity;
	u8 txDiversity;
	u16 txPower;
#define TXPOWER_DEFAULT 0
	u16 rssiThreshold;
#define RSSI_DEFAULT 0
        u16 modulation;
	u16 radioSpecific[3];
	/*---------- Aironet Extensions ----------*/
	u8 nodeName[16];
	u16 arlThreshold;
	u16 arlDecay;
	u16 arlDelay;
	u16 _reserved4[1];
	/*---------- Aironet Extensions ----------*/
	u16 magicAction;
#define MAGIC_ACTION_STSCHG 1
#define MACIC_ACTION_RESUME 2
#define MAGIC_IGNORE_MCAST (1<<8)
#define MAGIC_IGNORE_BCAST (1<<9)
#define MAGIC_SWITCH_TO_PSP (0<<10)
#define MAGIC_STAY_IN_CAM (1<<10)
	u16 filler;
} ConfigRid;

typedef struct {
	u16 len;
	u8 mac[6];
	u16 mode;
	u16 errorCode;
	u16 sigQuality;
	u16 SSIDlen;
	char SSID[32];
	char apName[16];
	char bssid[4][6];
	u16 beaconPeriod;
	u16 dimPeriod;
	u16 atimDuration;
	u16 hopPeriod;
	u16 channelSet;
	u16 channel;
	u16 hopsToBackbone;
	u16 apTotalLoad;
	u16 generatedLoad;
	u16 accumulatedArl;
	u16 signalQuality;
	u16 currentXmitRate;
	u16 apDevExtensions;
	u16 normalizedSignalStrength;
	u16 _reserved1;
	u8 apIP[4];
	u16 _reserved[7];
} StatusRid;

typedef struct {
	u16 len;
	char oui[3];
	u16 prodNum;
	char manName[32];
	char prodName[16];
	char prodVer[8];
	char factoryAddr[6];
	char aironetAddr[6];
	u16 radioType;
	u16 country;
	char callid[6];
	char supportedRates[8];
	char rxDiversity;
	char txDiversity;
	u16 txPowerLevels[8];
	u16 hardVer;
	u16 hardCap;
	u16 tempRange;
	u16 softVer;
	u16 softSubVer;
	u16 interfaceVer;
	u16 softCap;
	u16 bootBlockVer;
	u16 _reserved;
} CapabilityRid;

#ifndef __FreeBSD__
#ifdef SIOCIWFIRSTPRIV
#ifdef SIOCDEVPRIVATE
#define AIROOLDIOCTL	SIOCDEVPRIVATE
#define AIROOLDIDIFC 	AIROOLDIOCTL + 1
#endif /* SIOCDEVPRIVATE */
#else /* SIOCIWFIRSTPRIV */
#define SIOCIWFIRSTPRIV SIOCDEVPRIVATE
#endif /* SIOCIWFIRSTPRIV */
#define AIROIOCTL	SIOCIWFIRSTPRIV
#define AIROIDIFC 	AIROIOCTL + 1
#else /* __FreeBSD__ */
#define AIROIOCTL	SIOCGPRIVATE_0
#define AIROIDIFC 	SIOCGPRIVATE_1
#endif /* __FreeBSD__ */

#define AIROMAGIC	0xa55a

#define	AIROGCAP  		0	// Capability rid
#define AIROGCFG		1       // USED A LOT 
#define AIROGSLIST		2	// System ID list 
#define AIROGVLIST		3       // List of specified AP's
#define AIROGDRVNAM		4	//  NOTUSED
#define AIROGEHTENC		5	// NOTUSED
#define AIROGWEPKTMP		6
#define AIROGWEPKNV		7
#define AIROGSTAT		8
#define AIROGSTATSC32		9
#define AIROGSTATSD32		10

typedef struct aironet_ioctl {
  unsigned short command;	// What to do
  unsigned short len;		// Len of data
  unsigned char *data;		// d-data
} aironet_ioctl;

unsigned char buf[1024];

int
main(int	argc,
     char **	argv)
{
  int i,j,command,skfd = -1;		/* generic raw socket desc.	*/
#ifndef __FreeBSD__
  struct iwreq wrq;
#else
  struct ifreq wrq;
#endif
  aironet_ioctl ai;

  /* Create a channel to the NET kernel. */
  if((skfd = socket_open()) < 0)
    {
      perror("socket");
      exit(-1);
    }

  /* Special case for help... */
  if((argc != 3) || (!strncmp(argv[1], "-h", 9)) ||
     (!strcmp(argv[1], "--help")))
    {
      close(skfd);
      fprintf(stderr, "Usage: airorid interface rid\n");
      fprintf(stderr, "          0 = Capabilities\n");
      fprintf(stderr, "          1 = Config\n");
      fprintf(stderr, "          2 = SSIDs\n");
      fprintf(stderr, "          3 = AP List\n");
      fprintf(stderr, "          4 = Driver Name\n");
      fprintf(stderr, "          5 = Ethernet Encap\n");
      fprintf(stderr, "          6 = WEP Key temp\n");
      fprintf(stderr, "          7 = WEP Key perm\n");
      fprintf(stderr, "          8 = Status\n");
      fprintf(stderr, "          9 = Stats\n");
      fprintf(stderr, "         10 = Delta stats\n\n");
      exit(0);
    }

  strcpy (wrq.ifr_name, argv[1]);
  WRQ_DATA(wrq) = (caddr_t)&ai;
  ai.len = sizeof (buf);
  ai.data = buf;

  if((ioctl(skfd, AIROIDIFC, &wrq) < 0)
#ifdef AIROOLDIDIFC
  && (ioctl(skfd, AIROOLDIDIFC, &wrq) < 0)
#endif
  ) {
    perror("AIROIDIFC");
    close(skfd);
    return(-1);
  }
  if (*(int *)buf != AIROMAGIC) {
    fprintf (stderr, "Aironet extensions not found\n");
    close(skfd);
    return(-1);
  }

  strcpy (wrq.ifr_name, argv[1]);
  WRQ_DATA(wrq) = (caddr_t)&ai;
  command = atoi(argv[2]);
  ai.command = command;
  ai.len = sizeof (buf);
  ai.data = buf;

  if((ioctl(skfd, AIROIOCTL, &wrq) < 0)
#ifdef AIROOLDIOCTL
  && (ioctl(skfd, AIROOLDIOCTL, &wrq) < 0)
#endif
  ) {
    perror("AIROIOCTL");
    close(skfd);
    return(-1);
  }

  /* Close the socket. */
  close(skfd);

  printf ("Length: %d\n", ((unsigned short *)buf)[0]);
  i = 2;
  switch (command) {
  case 0:
    printf ("OUI: %d, %d, %d\n", buf[2], buf[3], buf[4]);
    printf ("Product Number: %d\n", *(unsigned short *)(buf + 6));
    printf ("Manufacturer Name: %.32s\n", buf + 8);
    printf ("Product Name: %.16s\n", buf + 40);
    printf ("Product Version: %.8s\n", buf + 56);
    printf ("Factory Address: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[64], buf[65], buf[66], buf[67], buf[68], buf[69]);
    printf ("Aironet Address: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[70], buf[71], buf[72], buf[73], buf[74], buf[75]);
    printf ("Radio Type: %d\n", *(unsigned short *)(buf + 76));
    if (*(unsigned short *)(buf + 78) > 7)
      printf ("Country: %d\n", *(unsigned short *)(buf + 78));
    else
      printf ("Country: %s\n", country[*(unsigned short *)(buf + 78)]);
    printf ("Call ID: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[80], buf[81], buf[82], buf[83], buf[84], buf[85]);
    printf ("Supported Rates:");
    j = 1;
    for (i = 86; i < 94; i++)
      if (buf[i]) {
	printf ("%s %d%s", j ? "" : ",", buf[i] / 2, (buf[i] & 1) ? ".5" : "");
	j = 0;
      }
    printf (" Mbps\n");
    printf ("RX Diversity: %d\n", *(unsigned char *)(buf + 94));
    printf ("TX Diversity: %d\n", *(unsigned char *)(buf + 95));
    printf ("TX Power Levels:");
    j = 1;
    for (i = 96; i < 112; i += 2)
      if (*(unsigned short *)(buf + i)) {
	printf ("%s %d", j ? "" : ",", *(unsigned short *)(buf + i));
	j = 0;
      }
    printf (" mW\n");
    printf ("Hardware Version: %d\n", *(unsigned short *)(buf + 112));
    printf ("Hardware Capabilities: %04X", *(unsigned short *)(buf + 114));
    if ((*(unsigned short *)(buf + 114)) & 1)
      printf (", CCK supported");
    if ((*(unsigned short *)(buf + 114)) & 2)
      printf (", Rev D hardware");
    if ((*(unsigned short *)(buf + 114)) & 4)
      printf (", High speed WEP");
    if ((*(unsigned short *)(buf + 114)) & 8)
      printf (", Short preamble");
    if ((*(unsigned short *)(buf + 114)) & 0x10)
      printf (", 128-bit WEP");
    if ((*(unsigned short *)(buf + 114)) & 0xFFE0)
      printf (", 0x%04X", (*(unsigned short *)(buf + 114)) & 0xFFE0);
    printf ("\n");
    printf ("Temperature Range: %d\n", *(unsigned short *)(buf + 116));
    printf ("Software Version: %X.%X\n", *(unsigned short *)(buf + 118), *(unsigned short *)(buf + 120));
    printf ("Interface Version: %X\n", *(unsigned short *)(buf + 122));
    printf ("Software Capabilities: %04X", *(unsigned short *)(buf + 124));
    if ((*(unsigned short *)(buf + 124)) & 1)
      printf (", Enhanced PSP");
    if ((*(unsigned short *)(buf + 124)) & 2)
      printf (", WEP support");
    if ((*(unsigned short *)(buf + 124)) & 4)
      printf (", Signal Quality support");
    if ((*(unsigned short *)(buf + 124)) & 8)
      printf (", Normalized RSSI support");
    if ((*(unsigned short *)(buf + 124)) & 0x10)
      printf (", Serial Deep Sleep support");
    if ((*(unsigned short *)(buf + 124)) & 0x20)
      printf (", Site Survey Mode support");
    if ((*(unsigned short *)(buf + 124)) & 0x40)
      printf (", WEP Auto Rate adaptation");
    if ((*(unsigned short *)(buf + 124)) & 0x80)
      printf (", 4 WEP key support");
    if ((*(unsigned short *)(buf + 124)) & 0x100)
      printf (", 128-bit WEP support");
    if ((*(unsigned short *)(buf + 124)) & 0xFE00)
      printf (", 0x%04X", (*(unsigned short *)(buf + 124)) & 0xFE00);
    printf ("\n");
    printf ("Boot Block Version: %X\n", *(unsigned short *)(buf + 126));
    printf ("Required Hardware: %X\n", *(unsigned short *)(buf + 128));
    i = 130;
    break;
  case 1:
    i = *(unsigned short *)(buf + 2);
    if (i & MODE_LLC_HOST)
      printf ("Mode: LLC, ");
    else
      printf ("Mode: Ethernet, ");
    if (i & MODE_AIRONET_EXTEND)
      printf ("Aironet extensions, ");
    if (i & MODE_AP_INTERFACE)
      printf ("AP extensions, ");
    if (i & 0x800)
      printf ("Antenna alignment, ");
    if (i & 0x1000)
      printf ("Ethernet LLC, ");
    if (i & 0x2000)
      printf ("Leaf node bridge, ");
    if (i & 0x4000)
      printf ("CF-Pollable, ");
    if (i & 0x8000)
      printf ("0x8000, ");
    switch (i & 0xFF) {
    case MODE_STA_IBSS: printf ("IBSS\n"); break;
    case MODE_STA_ESS: printf ("ESS\n"); break;
    case MODE_AP: printf ("AP\n"); break;
    case MODE_AP_RPTR: printf ("AP Repeater\n"); break;
    default: printf ("%d\n", i & 0xFF); break;
    }
    i = *(unsigned short *)(buf + 4);
    printf ("Rx Mode: ");
    if (i & RXMODE_DISABLE_802_3_HEADER)
      printf ("No 802.3 head, ");
    if (i & 0x200)
      printf ("Return normalized RSSI, ");
    if (i & 0xFC00)
      printf ("0x%04X, ", i & 0xFC00);
    switch (i & 0xFF) {
    case RXMODE_BC_MC_ADDR: printf ("Multicast and broadcast\n"); break;
    case RXMODE_BC_ADDR: printf ("No multicast\n"); break;
    case RXMODE_ADDR: printf ("No multicast or broadcast\n"); break;
    case RXMODE_RFMON: printf ("Wireless monitor\n"); break;
    case RXMODE_RFMON_ANYBSS: printf ("Wireless monitor - Any BSS\n"); break;
    case RXMODE_LANMON: printf ("LAN monitor\n"); break;
    default: printf ("%d\n", i & 0xFF); break;
    }
    printf ("Fragmentation Threshold: %d\n", *(unsigned short *)(buf + 6));
    printf ("RTS Threshold: %d\n", *(unsigned short *)(buf + 8));
    printf ("MAC: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[10], buf[11], buf[12], buf[13], buf[14], buf[15]);
    printf ("Rates:");
    j = 1;
    for (i = 16; i < 24; i++)
      if (buf[i]) {
	printf ("%s %d%s", j ? "" : ",", buf[i] / 2, (buf[i] & 1) ? ".5" : "");
	j = 0;
      }
    printf (" Mbps\n");
    printf ("Short Retry Limit: %d\n", *(unsigned short *)(buf + 24));
    printf ("Long Retry Limit: %d\n", *(unsigned short *)(buf + 26));
    printf ("TX Lifetime: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 28));
    printf ("RX Lifetime: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 30));
    printf ("Stationary: %d\n", *(unsigned short *)(buf + 32));
    printf ("Ordering: %d\n", *(unsigned short *)(buf + 34));
    printf ("Device Type: %04X\n", *(unsigned short *)(buf + 36));
    printf ("Contention Free Protocol Rate: %d\n", *(unsigned short *)(buf + 38));
    printf ("Contention Free Protocol Duration: %d\n", *(unsigned short *)(buf + 40));
    printf ("Reserved1: %d\n", *(unsigned short *)(buf + 42));
    printf ("Reserved2: %d\n", *(unsigned short *)(buf + 44));
    printf ("Reserved3: %d\n", *(unsigned short *)(buf + 46));
    i = *(unsigned short *)(buf + 48);
    switch (i) {
    case SCANMODE_ACTIVE: printf ("Scan Mode: Active\n"); break;
    case SCANMODE_PASSIVE: printf ("Scan Mode: Passive\n"); break;
    case SCANMODE_AIROSCAN: printf ("Scan Mode: Airoscan\n"); break;
    default: printf ("Scan Mode: %d\n", i); break;
    }
    printf ("Probe Delay: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 50));
    printf ("Probe Energy Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 52));
    printf ("Probe Response Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 54));
    printf ("Beacon Listen Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 56));
    printf ("Join Net Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 58));
    printf ("Auth Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 60));
    i = *(unsigned short *)(buf + 62);
    printf ("Auth Type: %04X", i);
    if (i & 0x100)
      printf (", Privacy in use");
    if (i & 0x200)
      printf (", Allow unencrypted traffic");
    if (i & 0xFC00)
      printf (", 0x%04X", i & 0xFC00);
    switch (i & 0x1FF) {
    case AUTH_OPEN: printf (", Open"); break;
    case AUTH_ENCRYPT: printf (", Encrypt"); break;
    case AUTH_SHAREDKEY: printf (", Shared Key"); break;
    default: printf (", %d", i & 0x1FF);
    }
    printf ("\n");
    printf ("Association Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 64));
    printf ("Specified AP Timeout: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 66));
    printf ("Offline Scan Interval: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 68));
    printf ("Offline Scan Duration: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 70));
    printf ("Link Loss Delay: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 72));
    printf ("Max Beacon Lost Time: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 74));
    printf ("Refresh Interval: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 76));
    printf ("Reserved1a: %d\n", *(unsigned short *)(buf + 78));
    i = *(unsigned short *)(buf + 80);
    switch (i) {
    case POWERSAVE_CAM: printf ("Power Save Mode: CAM\n"); break;
    case POWERSAVE_PSP: printf ("Power Save Mode: PSP\n"); break;
    case POWERSAVE_PSPCAM: printf ("Power Save Mode: PSPCAM\n"); break;
    default: printf ("Power Save Mode: %d\n", i); break;
    }
    printf ("Sleep For Dtims: %d\n", *(unsigned short *)(buf + 82));
    printf ("Listen Interval: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 84));
    printf ("Fast Listen Interval: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 86));
    printf ("Listen Decay: %d\n", *(unsigned short *)(buf + 88));
    printf ("Fast Listen Delay: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 90));
    printf ("Reserved1: %d\n", *(unsigned short *)(buf + 92));
    printf ("Reserved2: %d\n", *(unsigned short *)(buf + 94));
    printf ("Beacon Period: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 96));
    printf ("ATIM Duration: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 98));
    printf ("Reserved: %d\n", *(unsigned short *)(buf + 100));
    printf ("DS Channel: %d\n", *(unsigned short *)(buf + 102));
    printf ("Reserved: %d\n", *(unsigned short *)(buf + 104));
    printf ("DTIM Period: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 106));
    printf ("Bridge Distance: %d km\n", *(unsigned short *)(buf + 108));
    printf ("Radio ID: %d\n", *(unsigned short *)(buf + 110));
    i = *(unsigned short *)(buf + 112);
    switch (i) {
    case RADIOTYPE_DEFAULT: printf ("Radio Type: default\n"); break;
    case RADIOTYPE_802_11: printf ("Radio Type: 802.11\n"); break;
    case RADIOTYPE_LEGACY: printf ("Radio Type: legacy\n"); break;
    default: printf ("Radio Type: %d\n", i); break;
    }
    printf ("RX Diversity: %d\n", *(unsigned char *)(buf + 114));
    printf ("TX Diversity: %d\n", *(unsigned char *)(buf + 115));
    printf ("TX Power: %d mW\n", *(unsigned short *)(buf + 116));
    printf ("RSSI Threshold: %d\n", *(unsigned short *)(buf + 118));
    i = *(unsigned short *)(buf + 120);
    switch (i) {
    case 1: printf ("Modulation: CCK\n"); break;
    case 2: printf ("Modulation: MOK\n"); break;
    default: printf ("Modulation: %d\n", i); break;
    }
    i = *(unsigned short *)(buf + 122);
    switch (i) {
    case 0: printf ("Short Preamble: default\n"); break;
    case 1: printf ("Short Preamble: off\n"); break;
    case 2: printf ("Short Preamble: on\n"); break;
    default: printf ("Short Preamble: %d\n", i); break;
    }
    i = *(unsigned short *)(buf + 124);
    switch (i) {
    case 1: printf ("Home Product: Home Product\n"); break;
    case 2: printf ("Home Product: Install AP\n"); break;
    default: printf ("Home Product: %d\n", i); break;
    }
    printf ("Radio Specific1: %d\n", *(unsigned short *)(buf + 126));
    printf ("Node Name: \"%.16s\"\n", buf + 128);
    printf ("ARL Threshold: %d\n", *(unsigned short *)(buf + 144));
    printf ("ARL Decay: %d\n", *(unsigned short *)(buf + 146));
    printf ("ARL Delay: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 148));
    printf ("Reserved1: %d\n", *(unsigned short *)(buf + 150));
    i = *(unsigned char *)(buf + 152);
    printf ("Magic Action: ");
    if (i & MAGIC_IGNORE_MCAST)
      printf ("No multicast, ");
    if (i & MAGIC_IGNORE_BCAST)
      printf ("No broadcast, ");
    if (i & MAGIC_STAY_IN_CAM)
      printf ("Stay in CAM, ");
    else
      printf ("Switch to PSP, ");
    if (i & 0xF800)
      printf ("0x%04X, ", i & 0xF800);
    switch (i & 0xFF) {
    case MAGIC_ACTION_STSCHG: printf ("Status Change\n"); break;
    case MACIC_ACTION_RESUME: printf ("Resume\n"); break;
    default: printf ("%d\n", i); break;
    }
    printf ("Magic Packet Control: %d\n", *(unsigned char *)(buf + 153));
    printf ("Auto Wake Mask: %d\n", *(unsigned short *)(buf + 154));
    i = 156;
    break;
  case 2:
    { int l = *(unsigned short *)(buf + i);
    printf ("Length SSID1: %d\n", l);
    i += 2;
    for (j = 0; j < l; i++, j++) {
      printf ("%02X%c", buf[i], ((j % 16) == 15) ? '\n' : ' ');
    }
    l = *(unsigned short *)(buf + i);
    printf ("Length SSID2: %d\n", l);
    i += 2;
    for (j = 0; j < l; i++, j++) {
      printf ("%02X%c", buf[i], ((j % 16) == 15) ? '\n' : ' ');
    }
    l = *(unsigned short *)(buf + i);
    printf ("Length SSID3: %d\n", l);
    i += 2;
    for (j = 0; j < l; i++, j++) {
      printf ("%02X%c", buf[i], ((j % 16) == 15) ? '\n' : ' ');
    }}
    break;
  case 3:
    printf ("MAC1: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[2], buf[3], buf[4], buf[5], buf[6], buf[7]);
    printf ("MAC2: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[8], buf[9], buf[10], buf[11], buf[12], buf[13]);
    printf ("MAC3: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[14], buf[15], buf[16], buf[17], buf[18], buf[19]);
    printf ("MAC4: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[20], buf[21], buf[22], buf[23], buf[24], buf[25]);
    i = 26;
    break;
  case 5:
    do {
      if (buf[i] == 0 && buf[i+1] == 0)
	printf ("Ethertype: default   Action: ");
      else
	printf ("Ethertype: 0x%02X%02X    Action: ", buf[i], buf[i+1]);
      if (buf[i+2] & 1)
	printf ("keep RFC1042 in reception, ");
      if (buf[i+2] & 2)
	printf ("transmit using 802.1H\n");
      else
	printf ("transmit using RFC1042\n");
      i += 4;
    } while (i < 34 && buf[i-4] && buf[i-3]);
    break;
  case 6:
  case 7:
    printf ("Key index: %d\n", ((unsigned short *)buf)[1]);
    printf ("MAC: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[4], buf[5], buf[6], buf[7], buf[8], buf[9]);
    printf ("Key length: %d\n", ((unsigned short *)buf)[5]);
    i = 12;
    break;
  case 8:
    printf ("MAC: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[2], buf[3], buf[4], buf[5], buf[6], buf[7]);
    i = *(unsigned short *)(buf + 8);
    printf ("Mode: %04X", i);
    if (i & 1)
      printf (", Configured");
    if (i & 2)
      printf (", MAC Enabled");
    if (i & 4)
      printf (", Receive Enabled");
    if (i & 0x10)
      printf (", In Sync");
    if (i & 0x20)
      printf (", Associated");
    if (i & 0x8000)
      printf (", Error");
    if (i & 0x7FC8)
      printf (", 0x%04X", i & 0x7FC8);
    printf ("\n");
    printf ("Error Code: %d\n", *(unsigned short *)(buf + 10));
    printf ("Signal Quality: %d\n", *(unsigned short *)(buf + 12));
    printf ("SSID Length: %d\n", *(unsigned short *)(buf + 14));
    printf ("SSID: \"%.32s\"\n", buf + 16);
    printf ("AP Name: \"%.16s\"\n", buf + 48);
    printf ("BSSID1: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[64], buf[65], buf[66], buf[67], buf[68], buf[69]);
    printf ("BSSID2: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[70], buf[71], buf[72], buf[73], buf[74], buf[75]);
    printf ("BSSID3: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[76], buf[77], buf[78], buf[79], buf[80], buf[81]);
    printf ("BSSID4: %02X:%02X:%02X:%02X:%02X:%02X\n", buf[82], buf[83], buf[84], buf[85], buf[86], buf[87]);
    printf ("Beacon Period: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 88));
    printf ("DTIM Period: %d\n", *(unsigned short *)(buf + 90));
    printf ("ATIM Duration: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 92));
    printf ("Hop Period: %d usec\n", 1024 * (int)*(unsigned short *)(buf + 94));
    printf ("Channel Set: %d\n", *(unsigned short *)(buf + 96));
    printf ("Channel: %d\n", *(unsigned short *)(buf + 98));
    printf ("Hops to Backbone: %d\n", *(unsigned short *)(buf + 100));
    printf ("AP Total Load: %d\n", *(unsigned short *)(buf + 102));
    printf ("Generated Load: %d\n", *(unsigned short *)(buf + 104));
    printf ("Accumulated Arl: %d\n", *(unsigned short *)(buf + 106));
    printf ("Signal Quality: %d\n", *(unsigned short *)(buf + 108));
    i = *(unsigned short *)(buf + 110);
    printf ("Current Xmit Rate: %d%s Mbps\n", i / 2, (i & 1) ? ".5" : "");
    printf ("AP Dev Extensions: %d\n", *(unsigned short *)(buf + 112));
    printf ("Signal Strength: %d%%\n", *(unsigned short *)(buf + 114));
    printf ("Reserved1: %d\n", *(unsigned short *)(buf + 116));
    printf ("SSID IP: %d.%d.%d.%d\n", *(unsigned char *)(buf + 118),
	*(unsigned char *)(buf + 119), *(unsigned char *)(buf + 120),
	*(unsigned char *)(buf + 121));
    printf ("Reserved4: %d\n", *(unsigned short *)(buf + 122));
    printf ("Reserved5: %d\n", *(unsigned short *)(buf + 124));
    printf ("Reserved6: %d\n", *(unsigned short *)(buf + 126));
    printf ("Reserved7: %d\n", *(unsigned short *)(buf + 128));
    printf ("Reserved8: %d\n", *(unsigned short *)(buf + 130));
    printf ("Reserved9: %d\n", *(unsigned short *)(buf + 132));
    printf ("Reserved10: %d\n", *(unsigned short *)(buf + 134));
    i = 136;
    break;
  case 9:
  case 10:
    for (i = 0; statsLabels[i] != (char*)-1; i++)
      printf ("%s: %u\n", statsLabels[i], ((int *)buf)[i + 1]);
    i = 2 + (i << 2);
    break;
  }
  for (j = 0; i < *(unsigned short *)buf; i++, j++) {
    printf ("%02X%c", buf[i], ((j % 16) == 15) ? '\n' : ' ');
  }
  printf ("\n");

  return(0);
}
